"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from typing import no_type_check

import torch
from torch import Tensor
from torch.functional import einsum
from torch.nn import Module
from torch.nn.functional import kl_div


class AssociatorLoss(Module):
    """
    probabilistic associator loss
    """
    # pylint: disable=arguments-differ
    @no_type_check
    def forward(self, cayley_cube: Tensor) -> Tensor:
        """
        finds a probabilistic associator of a given probabilistic Cayley cube

:math:`P\\left\\{e_i\\left(e_je_k\\right)=e_l\\right\\}=
\\sum\\limits_{m=1}^nP\\left\\{e_i\\left(e_je_k\\right)=e_l\\vert
e_je_k=e_m\\right\\}P\\left\\{e_je_k=e_m\\right\\}=
\\sum\\limits_{m=1}^na_{iml}a_{jkm}`

        :param cayley_cube: this is a probabilistic Cayley cube
        :returns: the probabilistic associator
        """
        one = einsum("biml,bjkm->bijkl", cayley_cube, cayley_cube)
        two = einsum("bmkl,bijm->bijkl", cayley_cube, cayley_cube)
        return (kl_div(torch.log(one), two, reduction="sum") /
                cayley_cube.shape[0])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neural_semigroups']

package_data = \
{'': ['*']}

install_requires = \
['numpy==1.18.1',
 'pandas==1.0.2',
 'pytorch-ignite==0.3.0',
 'tensorboard==2.1.1',
 'torch==1.4.0',
 'tqdm==4.43.0']

setup_kwargs = {
    'name': 'neural-semigroups',
    'version': '0.2.3',
    'description': 'Neural networks powered research of semigroups',
    'long_description': "# Neural Semigroups\n\nHere we try to model Cayley tables of semigroups using neural networks.\n\nMore documentation can be found \n[here](https://neural-semigroups.readthedocs.io).\n\n## Motivation\n\nThis work was inspired by [a sudoku\nsolver](https://github.com/Kyubyong/sudoku). A solved Sudoku puzzle\nis nothing more than a Cayley table of a quasigroup from 9 items with\nsome well-known additional properties. So, one can imagine a puzzle\nmade from a Cayley table of any other magma, e. g. a semigroup, by\nhiding part of its cells.\n\nThere are two major differences between sudoku and puzzles based on\nsemigroups:\n\n1) it's easy to take a glance on a table to understand whether it is\na sudoku or not. That's why it was possible to encode numbers in a\ntable cells as colour intensities. Sudoku is a picture, and a\nsemigroup is not. It's difficult to check a Cayley table's\nassociativity with a naked eye;\n\n2) sudoku puzzles are solved by humans for fun and thus catalogued.\nWhen solving a sudoku one knows for sure that there is a unique\nsolution. On the contrary, nobody guesses values in a partially\nfilled Cayley table of a semigroup as a form of amuzement. As a\nresult, one can create a puzzle from a full Cayley table of a\nsemigroup but there may be many distinct solutions.\n",
    'author': 'Boris Shminke',
    'author_email': 'boris@shminke.ml',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inpefess/neural-semigroups',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np


class Magma:
    """
    an implementation of `a magma`_ (a set with a binary operation)

    .. _a magma: https://en.wikipedia.org/wiki/Magma_%28algebra%29
    """
    cayley_table: np.ndarray

    def __init__(self, cayley_table: np.ndarray):
        """
        :param cayley_table: a Cayley table for a magma
        """
        self.cayley_table = cayley_table

    @property
    def cardinality(self) -> int:
        """
        number of elements in a magma
        """
        return self.cayley_table.shape[0]

    @property
    def is_associative(self) -> bool:
        """
        check associativity of a Cayley table

        :returns: whether the input table is assosiative or not

        """
        a_range = range(self.cardinality)
        for one in a_range:
            for two in a_range:
                if (
                        self.cayley_table[
                            self.cayley_table[one, two], a_range
                        ] !=
                        self.cayley_table[
                            one, self.cayley_table[two, a_range]
                        ]
                ).any():
                    return False
        return True

    @property
    def is_commutative(self) -> bool:
        """
        check commutativity of a Cayley table

        :returns: whether the input table is commutative or not

        """
        return np.allclose(
            self.cayley_table, self.cayley_table.T
        )

    @property
    def identity(self) -> int:
        """
        find an identity element in a Cayley table

        :returns: the index of the identity element or -1 if there is no identity
        """
        identity_row = np.arange(self.cardinality)
        identity = -1
        for i in identity_row:
            if np.allclose(self.cayley_table[i, :], identity_row):
                identity = i
                break
        return identity

    @property
    def has_inverses(self) -> bool:
        """
        check whether there are solutions of equations :math:`ax=b` and :math:`xa=b``
        """
        identity_row = np.arange(self.cardinality)
        has_inverses = True
        for i in identity_row:
            if (
                    not np.allclose(
                        np.sort(self.cayley_table[i, :]), identity_row
                    ) or
                    not np.allclose(
                        np.sort(self.cayley_table[:, i]), identity_row
                    )
            ):
                has_inverses = False
                break
        return has_inverses

    @property
    def probabilistic_cube(self) -> np.ndarray:
        """
        a 3d array :math:`a` where :math:`a_{ijk}=P\\left\\{e_ie_j=e_k\\right\\}`

        :returns: a probabilistic cube representation of a Cayley table
        """
        cube = np.zeros(
            [self.cardinality, self.cardinality, self.cardinality],
            dtype=np.float32
        )
        for i in range(self.cardinality):
            for j in range(self.cardinality):
                cube[i, j, self.cayley_table[i, j]] = 1.0
        return cube
